/*
* Copyright (c) 2003, University of Maryland 
* Copyright (c) 2003, Center for Satellite and Hybrid Communication Networks 
* All rights reserved.
*
* Redistribution and use in source and binary forms, with or without 
* modification, are permitted provided that the following conditions 
* are met:
*
*   * Redistributions of source code must retain the above copyright notice, 
*     this list of conditions,the authors and the following disclaimer.
*   * Redistributions in binary form must reproduce the above copyright notice,
*     this list of conditions,the authors and the following disclaimer in
*     the documentation and/or other materials provided with the distribution.
*   * Neither the name of the University of Maryland/Center for Satellite
*     and Hybrid Communication Networks nor the names of its 
*     contributors may be used to endorse or promote products derived from 
*     this software without specific prior written permission.
*
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
* AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
* IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
* ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
* LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
* CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
* SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
* INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
* ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
* POSSIBILITY OF SUCH DAMAGE.
*
*/

/* Authors: 
*     Jonathan Polley (jpolley@isr.umd.edu),
*     Dionysus Blazakis (dblaze@isr.umd.edu),
*     Jonathan McGee (mcgee@isr.umd.edu),
*     Daniel Rusk (ruskd@isr.umd.edu),
*     Manish Karir (karir@isr.umd.edu)
*/

#include <assert.h>
#include <math.h>
#include "radio.h"
/*
 * This module handles a basic transmission model for the atemu's MICA2
 * radio system.  Each node will have a pointer to the air, and interface
 * other devices by passing this structure to these functions.
 *
 * Limitations:
 * - The radio has an infinitely precise tuner.
 * - Behavior at a range of '0' is...well...physically impossible
 * - Antenna length is not taken into effect.
 * - Radio will only transmit one frequency at a time.
 * - The radio can only receive a frequency modulated binary value, amplitude
 *   modulation can be faked with the strength parameter on air_read_bit().
 *
 * Assumptions:
 * - The transmission power is on an absolute, linear scale.  Overlapping
 *   transmissions can be added.
 * - The received power decreases with the square of distance
 *
 * Best practices:
 * - Use Watts or Milliwatts for transmitting power
 * - Use Watts/m^2 or Milliwatts/m^2 for detection threshold
 * - Use meters for positions
 */

/* This increments every time someone messes with the air waves */
static int32_t currentcycle = 0;
/* All the air devices are stored in a linked list */
static radio_air *firstdevice = NULL;
static radio_air *lastdevice = NULL;
/* Number of devices in the air */
static int devicecount = 0;

radio_air* air_create_device() {
    radio_air* current;
    radio_aircache *cache, *prev = NULL;
    double distance;

    /* Create initial device */
    radio_air* newnode = (radio_air*)malloc(sizeof(radio_air));
    if(newnode == NULL) {
        return NULL;
    }

    /* Initialize with "sane" values */
    newnode->x = 0;
    newnode->y = 0;
    newnode->z = 0;
    newnode->transmitting = 0;
    newnode->lastchange = currentcycle;
    newnode->freqtrue = 1;
    newnode->freqfalse = 0;
    newnode->strengthtrue = 0.0;
    newnode->strengthfalse = 0.0;
    newnode->prev = NULL;
    newnode->next = NULL;
    newnode->cache = NULL;

    /* Create initial cache */
    current = firstdevice;
    while(current) {
        cache = (radio_aircache*)malloc(sizeof(radio_aircache));
        if(cache == NULL) {
            air_destroy_device(newnode);
            return NULL;
        }

        cache->remote = current;
        distance = pow(current->x, 2) + pow(current->y, 2) + pow(current->z, 2);
        if(distance == 0) {
            cache->divisor = 1;
        } else {
            cache->divisor = distance;
        }
        cache->frequency = current->frequency;
        if(current->transmitting) {
            cache->strength = current->strength / cache->divisor;
        } else {
            cache->strength = 0;
        }
        cache->lastchange = currentcycle;
        cache->next = NULL;

        if(prev) {
            prev->next = cache;
        } else {
            newnode->cache = cache;
        }
        prev = cache;
        current = current->next;
    }

    /* Add to cache of other pre-existing entries */
    current = firstdevice;
    while(current) {
        cache = (radio_aircache*)malloc(sizeof(radio_aircache));
        if(cache == NULL) {
            air_destroy_device(newnode);
            return NULL;
        }

        cache->remote = newnode;
        distance = pow(current->x, 2) + pow(current->y, 2) + pow(current->z, 2);
        if(distance == 0) {
            cache->divisor = 1;
        } else {
            cache->divisor = distance;
        }
        cache->strength = 0;
        cache->frequency = 0;
        cache->lastchange = currentcycle;

        cache->next = current->cache;
        current->cache = cache;
        current = current->next;
    }

    /* Add to the linked list */
    devicecount++;
    if(firstdevice == NULL) {
        /* First device created */
        firstdevice = lastdevice = newnode;
    } else {
        /* Devices already exist, add to end */
        newnode->prev = lastdevice;
        lastdevice->next = newnode;
        lastdevice = newnode;
    }

    return newnode;
}

error_t air_set_position(radio_air *node, double x, double y, double z) {
    radio_air *current;
    radio_aircache *cache;
    double distance;

    /* Quick test before wasting time */
    if((node->x == x) && (node->y == y) && (node->z == z)) {
        return 0;
    }

    /* Update our position and update the cycle count */
    node->x = x;
    node->y = y;
    node->z = z;
    if(node->transmitting) {
        node->lastchange = ++currentcycle;
    }

    /* Divisors are only recalculated on position changes */
    current = firstdevice;
    while(current) {
        if(current != node) {
            /* Find this node's cache entry for the moved node */
            cache = current->cache;
            while(cache && cache->remote != node) {
                cache = cache->next;
            }
            /* Update the new position */
            if(cache) {
                distance = pow(x - current->x, 2) + pow(y - current->y, 2)
                        + pow(z - current->z, 2);
                if(distance == 0) {
                    cache->divisor = 1;
                } else {
                    cache->divisor = distance;
                }
            }
        }
        current = current->next;
    }

    /* Fix our own distances */
    cache = node->cache;
    while(cache) {
        distance = pow(x - cache->remote->x, 2) + pow(y - cache->remote->y, 2)
                + pow(z - cache->remote->z, 2);
        if(distance == 0) {
            cache->divisor = 1;
        } else {
            cache->divisor = distance;
        }
        cache = cache->next;
    }

    /* Return success */
    return 0;
}

error_t air_destroy_device(radio_air *node) {
    radio_aircache *prev, *next;
    radio_air *current;

    /* Release its cache */
    prev = node->cache;
    while(prev) {
        next = prev->next;
        free(prev);
        prev = next;
    }
    node->cache = NULL;

    /* Release other nodes' cache of this node */
    current = firstdevice;
    while(current) {
        next = current->cache;
        prev = NULL;
        while(next && (next->remote != node)) {
            prev = next;
            next = next->next;
        }

        if(next) {
            if(prev) {
                prev->next = next->next;
            } else {
                current->cache = next->next;
            }
            free(next);
        }

        /* If the current node, remove from list */
        if(current == node) {
            if(current->prev) {
                current->prev->next = current->next;
            } else {
                firstdevice = current->next;
            }

            if(current->next) {
                current->next->prev = current->prev;
            } else {
                lastdevice = current->prev;
            }
        }

        current = current->next;
    }

    /* Free and return */
    node->next = NULL;
    node->prev = NULL;
    free(node);
    return 0;
}

/************************
 ** Transmit Functions **
 ************************/

/* set the transmission power */
error_t air_transmit_power(radio_air *node, double power){
    if(node->strength == power){
        return 0;
    }
    
    node->strength = power;
    node->lastchange = ++currentcycle;

    return 0;
}

/* set that we are transmitting such and such a frequency */
error_t air_transmit(radio_air *node, double freq){
    if(node->transmitting && node->frequency == freq){
        return 0;
    }

    node->transmitting = 1;
    node->frequency = freq;
    node->lastchange = ++currentcycle;

    return 0;
}

/* stop transmitting */
error_t air_transmit_release(radio_air *node){
    if(!(node->transmitting)){
        return 0;
    }

    node->transmitting=0;
    node->lastchange = ++currentcycle;

    return 0;
}

/************************
 ** Transmit Functions **
 ************************/

/* hint as to if there's something to receive */
int air_receive_changed(radio_air *node) {
    return (node->lastread != currentcycle);
}

/* set tuner params */
error_t air_receive_tuner(radio_air *node, double frequency, double deltafreq,
        double threshold){
    if(node->frequency == frequency && node->deltafreq == deltafreq && node->threshold == threshold){
        return 0;
    }

    node->frequency = frequency;
    node->deltafreq = deltafreq;
    node->threshold = threshold;
    node->lastchange = ++currentcycle;

    return 0;
}

/* If there has been a change, update the datastructures */
error_t air_cache_update(radio_air *node) {
    if(node->lastread != currentcycle){
        radio_aircache *cache = node->cache;

        node->lastread = currentcycle;

        while(cache) {
            /* Update this cache entry */
            if(cache->lastchange != cache->remote->lastchange) {
                if(cache->remote->transmitting) {
                    cache->frequency = cache->remote->frequency;
                    cache->strength = cache->remote->strength / cache->divisor;
                    cache->transmitting = 1;
                } else {
                    cache->transmitting = 0;
                }
                cache->lastchange = cache->remote->lastchange;
            }
            cache = cache->next;
        }
    }

    return 0;
}

/* runs tuning on node->frequency.  finds strength, and if frequency != NULL.
 * also fills in average frequency on the air */
error_t air_tune_freq(radio_air *node, double tunefreq, double *frequency,
        double *strength){
    radio_aircache *cache = node->cache;
    int activity = 0;

    if(frequency != NULL){
        *frequency=0;
    }
    *strength=0;

    /* count up weighted frequencies and signal strength */
    /* use tuned frequency and deltafreq to drop off more distant freqs */
    while(cache){
        if(cache->transmitting){
            double scalar;
            scalar = pow(10, - fabs(cache->frequency - tunefreq) / node->deltafreq);
            scalar *= cache->strength;
            *strength += scalar;
            if(frequency != NULL){
                *frequency += cache->frequency * scalar;
            }
            activity = 1;
        }
        cache=cache->next;
    }

    if((frequency != NULL) && activity) {
        *frequency /= *strength;
    }

    return 0;
}

/* try to tune to a frequency */
int air_receive_sense(radio_air *node, double *frequency, double *strength){
    air_cache_update(node);
   
    air_tune_freq(node, node->frequency, frequency, strength);

    /* if we found a good signal, find the weighted average */
    if(*strength > node->threshold){
        return 1;
    }
    
    /* no (suitable) transmissions */
    return 0;
}

/* set true and false frequencies, and transition threshold */
error_t air_receive_frequencies(radio_air *node, double freqfalse,
        double freqtrue, double schmittdist){
    if(node->freqtrue == freqtrue && node->freqfalse == freqfalse &&
            node->schmittdist == schmittdist){
        return 0;
    }

    node->freqtrue = freqtrue;
    node->freqfalse = freqfalse;
    node->schmittdist = schmittdist;
    node->lastchange = ++currentcycle;

    return 0;
}

int air_receive(radio_air *node, int *bitout, double *strengthout){
    air_cache_update(node);

    /* perform tuning on both truefreq and falsefreq */
    air_tune_freq(node, node->freqtrue, NULL, &(node->strengthtrue));
    air_tune_freq(node, node->freqfalse, NULL, &(node->strengthfalse));

    if(node->strengthtrue > 0 && node->strengthfalse > 0){
        /* must do Schmitt-ing */
        if(node->lastfreq == node->freqfalse){
            /* if we exceed threshold, become true */
            if((node->strengthtrue) / (node->strengthfalse + node->strengthtrue) > node->schmittdist){
                node->lastfreq = node->freqtrue;
                *bitout=1;
                *strengthout=node->strengthtrue;
                return (*strengthout > node->threshold);
            }else{
                /* we remain false */
                node->lastfreq = node->freqfalse;
                *bitout=0;
                *strengthout=node->strengthfalse;
                return (*strengthout > node->threshold);
            }
        }else{
            /* if we exceed threshold, become false */
            if((node->strengthfalse) / (node->strengthfalse + node->strengthtrue) > node->schmittdist){
                node->lastfreq = node->freqfalse;
                *bitout=0;
                *strengthout=node->strengthfalse;
                return (*strengthout > node->threshold);
            }else{
                /* we remain true */
                node->lastfreq = node->freqtrue;
                *bitout=1;
                *strengthout=node->strengthtrue;
                return (*strengthout > node->threshold);
            }
        }
    }else if(node->strengthtrue > 0){
        /* only true is strong enough */
        node->lastfreq = node->freqtrue;
        *bitout=1;
        *strengthout=node->strengthtrue;
        return (*strengthout > node->threshold);
    }else if(node->strengthfalse > 0){
        /* only false is strong enough */
        node->lastfreq = node->freqfalse;
        *bitout=0;
        *strengthout=node->strengthfalse;
        return (*strengthout > node->threshold);
    }else{
        /* nothing was > 0 */
        *bitout=0;
        *strengthout=0.0;
        return 0;
    }
}

